import os
import numpy as np
import math
import pandas as pd
import cv2
import pickle
import scipy.ndimage as ndi
import colorsys
import pyqtgraph as pg
import pyqtgraph.opengl as gl
from scipy.interpolate import interp1d, splprep, splev


def read_label(file):
    lines = []
    for line in file:
        lines.append(line)
    
    n_lines = len(lines)
    label_index = np.zeros(n_lines - 14, 'i')
    label_index[:] = np.nan
    label_colors = np.zeros((n_lines - 14, 3), 'i')
    label_colors[:] = np.nan
    label_names = []
    for i in range(n_lines - 14):
        split_lines = lines[i + 14].split()
        label_index[i] = int(split_lines[0])
        label_colors[i] = np.array([split_lines[1], split_lines[2], split_lines[3]]).astype(int)
        split_lines2 = lines[i + 14].split('"')
        label_names.append(split_lines2[1])
    
    return label_index, label_names, label_colors


def rotation_x(theta):
    ct = np.cos(theta)
    st = np.sin(theta)
    rx = np.array([[1, 0, 0], [0, ct, -st], [0, st, ct]])
    return rx


def rotation_y(theta):
    ct = np.cos(theta)
    st = np.sin(theta)
    ry = np.array([[ct, 0, st], [0, 1, 0], [-st, 0, ct]])
    return ry


def rotation_z(theta):
    ct = np.cos(theta)
    st = np.sin(theta)
    rz = np.array([[ct, -st, 0], [st, ct, 0], [0, 0, 1]])
    return rz



def d2td3(pos2d, ax, ay, o):
    pos3d = pos2d[0] * ax + pos2d[1] * ay + o
    return pos3d


def get_closest_point_to_line(p0, r, p):
    ap = p - p0
    res = p0 + np.dot(ap, r) * r
    return res
    

def line_fit(points):
    print('input', points)
    points = np.asarray(points)
    sort_order = np.argsort(points[:, 2])[::-1]
    points = points[sort_order, :]
    print('reorder', points)
    avg = np.mean(points, 0)
    substracted = points - avg
    u, s, vh = np.linalg.svd(substracted)
    direction = vh[0, :] / np.linalg.norm(vh[0, :])
    print('direct', direction)
    p1 = points[0, :]
    p2 = points[-1, :]
    sp = get_closest_point_to_line(avg, direction, p1)
    ep = get_closest_point_to_line(avg, direction, p2)
    return sp, ep, avg, direction


def get_angles(direction):
    direction = direction / np.linalg.norm(direction)
    
    vertical_vec = np.array([0, 0, 1])
    
    ap_proj = direction.copy()
    ap_proj[0] = 0
    ap_proj = ap_proj / np.linalg.norm(ap_proj)

    ml_proj = direction.copy()
    ml_proj[1] = 0
    ml_proj = ml_proj / np.linalg.norm(ml_proj)

    ap_val = math.acos(np.max([np.min([np.dot(ap_proj, vertical_vec), 1]), -1]))
    ml_val = math.acos(np.max([np.min([np.dot(ml_proj, vertical_vec), 1]), -1]))

    ap_angle = np.degrees(ap_val)
    ml_angle = np.degrees(ml_val)

    if ap_angle > 90:
        ap_angle = 180 - ap_angle
    if ml_angle > 90:
        ml_angle = 180 - ml_angle

    return ap_angle, ml_angle


def pandas_to_str(label_name, label_ano, length, channels):
    df = pd.DataFrame({
        'Brain Regions': label_name,
        'Ano': label_ano,
        'Probe Length': length,
        'Probe Channels': channels
        })
    return df.to_string(col_space=30, justify="justify")


def correct_start_pnt(label_data, start_pnt, start_vox, direction):
    direction = direction / np.linalg.norm(direction)
    check_vec = label_data[int(start_vox[0]), int(start_vox[1]), :]
    top_vox = np.where(check_vec != 0)[0][-1]
    if int(start_vox[2]) < top_vox:
        for i in range(1000):
            temp = start_vox - i * direction
            check_vox = temp.astype(int)
            if label_data[check_vox[0], check_vox[1], check_vox[2]] == 0:
                break
        if i == 999:
            print('something went wrong, please contact maintainer')
        new_sp = start_pnt - (i - 1) * direction
    elif int(start_vox[2]) > top_vox:
        for i in range(1000):
            temp = start_vox + i * direction
            check_vox = temp.astype(int)
            if label_data[check_vox[0], check_vox[1], check_vox[2]] == 0:
                break
        if i == 999:
            print('something went wrong, please contact maintainer')
        new_sp = start_pnt + (i - 1) * direction
    else:
        new_sp = start_pnt
    return new_sp


def correct_end_point(sp, ep, direction, vox_size, tip_length, probe_type):
    probe_length_with_tip = np.sqrt(np.sum((sp - ep) ** 2)) * vox_size  # in um
    probe_length_without_tip = probe_length_with_tip - tip_length  # in um
    if probe_type != 2:
        if probe_length_without_tip > 9600:
            probe_length_without_tip = 9600
            probe_length_with_tip = probe_length_without_tip + tip_length
        new_ep = sp + direction * probe_length_with_tip / vox_size
    else:
        new_ep = ep
    return new_ep, probe_length_with_tip, probe_length_without_tip


def check_parallel_to_z(sp, ep):
    d = ep - sp
    is_parallel = False
    if abs(d[0]) < 1e-6 and abs(d[1]) < 1e-6 and abs(d[2]) > 1:
        is_parallel = True
    return is_parallel


def angle_between_2vectors(vector1, vector2):
    unit_vector_1 = vector1 / np.linalg.norm(vector1)
    unit_vector_2 = vector2 / np.linalg.norm(vector2)
    dot_product = np.dot(unit_vector_1, unit_vector_2)
    angle = np.arccos(dot_product)
    return angle


def rotation_y(theta):
    ct = np.cos(theta)
    st = np.sin(theta)
    ry = np.array([[ct, 0, st], [0, 1, 0], [-st, 0, ct]])
    return ry


def rotation_z(theta):
    ct = np.cos(theta)
    st = np.sin(theta)
    rz = np.array([[ct, -st, 0], [st, ct, 0], [0, 0, 1]])
    return rz


def get_probe_info(probe_type):
    if probe_type == 0:
        tip_length = 175
        channel_size = 20
        channel_number_in_banks = (384, 384, 192)
    elif probe_type == 1:
        tip_length = 175
        channel_size = 15
        channel_number_in_banks = (384, 384, 192)
    else:
        tip_length = 0
        channel_size = None
        channel_number_in_banks = None
    return tip_length, channel_size, channel_number_in_banks


def get_probe_sites(segmentation_data, sp, ep, probe_length_without_tip,
                    direction, vox_size, probe_type, channel_size, bregma, site_face):
    if probe_type != 2:
        base_loc = np.array([-16, -8, 8, 16]) / vox_size

        center_loc = []
        da_loc = sp + direction * (probe_length_without_tip - 6) / vox_size

        while da_loc[2] < sp[2]:
            center_loc.append(da_loc)
            da_loc = center_loc[-1] - direction * channel_size / vox_size

        n_sites_row = len(center_loc)
        print(n_sites_row)

        sites_loc = []
        zero_vec = np.zeros(4)
        if check_parallel_to_z(sp, ep):
            if site_face == 0:
                # calculate for flat case
                related_site_center = np.vstack([base_loc, zero_vec, zero_vec]).T
                for i in range(n_sites_row):
                    sites_loc.append(center_loc[i] + related_site_center)
            else:
                # calculate for side case
                related_site_center = np.vstack([zero_vec, base_loc, zero_vec]).T
                for i in range(n_sites_row):
                    sites_loc.append(center_loc[i] + related_site_center)
        else:
            x_vec = np.array([1, 0, 0])
            vec1 = ep - sp
            vec2 = vec1.copy()
            vec2[2] = 0
            ang_alpha = angle_between_2vectors(vec2, x_vec)
            ang_beta = angle_between_2vectors(vec1, vec2)
            rot_m = np.dot(rotation_z(ang_alpha), rotation_y(ang_beta))

            if site_face == 0:
                # calculate for flat case
                related_site_center = np.vstack([zero_vec, base_loc, zero_vec])
                for i in range(n_sites_row):
                    sites_loc.append(center_loc[i] + np.dot(rot_m, related_site_center).T)
            else:
                # calculate for side case
                related_site_center = np.vstack([zero_vec, zero_vec, base_loc])
                for i in range(n_sites_row):
                    sites_loc.append(center_loc[i] + np.dot(rot_m, related_site_center).T)

        sites_label = np.zeros((n_sites_row, 2), 'i')
        for i in range(n_sites_row):
            if probe_type == 0:
                if np.mod(i, 2) == 0:
                    valid_sites_index = np.array([1, 3])
                else:
                    valid_sites_index = np.array([0, 2])
            else:
                valid_sites_index = np.array([1, 3])

            site_vox = sites_loc[i] + bregma
            site_vox = site_vox.astype(int)
            valid_vox = site_vox[valid_sites_index, :]
            sites_label[i, 0] = segmentation_data[valid_vox[0][0], valid_vox[0][1], valid_vox[0][2]]
            sites_label[i, 1] = segmentation_data[valid_vox[1][0], valid_vox[1][1], valid_vox[1][2]]

        region_label = np.unique(sites_label).astype(int)
        region_length = np.zeros(len(region_label))
        region_channels = np.zeros(len(region_label))
        for i in range(len(region_label)):
            region_channels[i] = np.sum(sites_label == region_label[i])
            region_length[i] = np.sum(sites_label == region_label[i]) * channel_size / 2
    else:
        sites_loc = None
        sites_label = None
        region_label = None
        region_length = None
        region_channels = None
        
    return sites_loc, sites_label, region_label, region_length, region_channels


def get_label_name(label_info, unique_label, sites_label):
    label_names = []
    label_acronym = []
    label_color = []
    left_sites_color = np.zeros((len(sites_label), 3), 'i')
    right_sites_color = np.zeros((len(sites_label), 3), 'i')
    for i in range(len(unique_label)):
        if unique_label[i] == 0:
            label_names.append(' ')
            label_acronym.append(' ')
            label_color.append((128, 128, 128))
            cind = np.where(sites_label[:, 0] == unique_label[i])[0]
            if len(cind) > 1:
                left_sites_color[cind, 0] = 128
                left_sites_color[cind, 1] = 128
                left_sites_color[cind, 2] = 128
            cind = np.where(sites_label[:, 1] == unique_label[i])[0]
            if len(cind) > 1:
                right_sites_color[cind, 0] = 128
                right_sites_color[cind, 1] = 128
                right_sites_color[cind, 2] = 128
        else:
            da_ind = np.where(np.ravel(label_info['index']) == unique_label[i])[0][0]
            label_names.append(label_info['label'][da_ind])
            label_acronym.append(label_info['abbrev'][da_ind])
            label_color.append(label_info['color'][da_ind])
            cind = np.where(sites_label[:, 0] == unique_label[i])[0]
            if len(cind) > 1:
                left_sites_color[cind, 0] = label_info['color'][da_ind][0]
                left_sites_color[cind, 1] = label_info['color'][da_ind][1]
                left_sites_color[cind, 2] = label_info['color'][da_ind][2]
            cind = np.where(sites_label[:, 1] == unique_label[i])[0]
            if len(cind) > 1:
                right_sites_color[cind, 0] = label_info['color'][da_ind][0]
                right_sites_color[cind, 1] = label_info['color'][da_ind][1]
                right_sites_color[cind, 2] = label_info['color'][da_ind][2]

    return label_names, label_acronym, label_color, left_sites_color, right_sites_color


def block_same_label_one_side(sites_label_one_side, sites_color_one_side):
    merged_labels = [sites_label_one_side[0]]
    merged_colors = [sites_color_one_side[0]]
    in_block_count = 1
    block_count = []
    for i in range(1, len(sites_label_one_side)):
        if sites_label_one_side[i] == sites_label_one_side[i - 1]:
            in_block_count += 1
        else:
            block_count.append(in_block_count)
            in_block_count = 1
            merged_labels.append(sites_label_one_side[i])
            merged_colors.append(sites_color_one_side[i])
        if i == len(sites_label_one_side) - 1:
            block_count.append(in_block_count)
    return merged_labels, merged_colors, block_count





def get_tilt_info(sp, ep):
    if ep[0] < sp[0]:
        ap_tilt = 'posterior'
    elif ep[0] > sp[0]:
        ap_tilt = 'anterior'
    else:
        ap_tilt = 'no tilt'

    if sp[1] > 0:
        if ep[1] < sp[1]:
            ml_tilt = 'medial'
        elif ep[1] > sp[1]:
            ml_tilt = 'lateral'
        else:
            ml_tilt = 'no tilt'
    elif sp[1] < 0:
        if ep[1] < sp[1]:
            ml_tilt = 'lateral'
        elif ep[1] > sp[1]:
            ml_tilt = 'medial'
        else:
            ml_tilt = 'no tilt'
    else:
        if ep[1] != sp[1]:
            ml_tilt = 'lateral'
        else:
            ml_tilt = 'no tilt'
    return ap_tilt, ml_tilt


def calculate_probe_info(data, label_data, label_info, vxsize_um, probe_type, bregma, site_face):
    """

    :param data: 3d coordinates for all the points
    :param label_data: original brain region segmentation
    :param label_info:
    :param vxsize_um:
    :param tip_length:
    :param channel_size:
    :param bregma:
    :return:
    """
    # find the best fit line of the given points
    # direction is from ???
    # start_pnt and end_pnt are coordinates related to the given Bregma
    tip_length, channel_size, channel_number_in_banks = get_probe_info(probe_type)
    start_pnt, end_pnt, avg, direction = line_fit(data)
    print(bregma)
    start_vox = start_pnt + bregma
    end_vox = end_pnt + bregma
    ap_angle, ml_angle = get_angles(direction)
    new_sp = correct_start_pnt(label_data, start_pnt, start_vox, direction)
    new_start_vox = new_sp + bregma
    print('sp', start_pnt)
    print('sp_vox', start_vox)
    print('new_sp', new_sp)
    print('new_sp_vox', new_start_vox)
    new_ep, probe_length_with_tip, probe_length_without_tip = correct_end_point(
        new_sp, end_pnt, direction, vxsize_um, tip_length, probe_type)
    print('ep', end_pnt)
    print('new_ep', new_ep)
    print(probe_length_with_tip)
    print(np.sqrt(np.sum((new_sp - new_ep) ** 2)) * vxsize_um)
    new_end_vox = new_ep + bregma
    new_end_vox = new_end_vox.astype(int)
    print(new_end_vox)

    sites_loc, sites_label, region_label, region_length, region_channels = get_probe_sites(
        label_data, new_sp, new_ep, probe_length_without_tip,
        direction, vxsize_um, probe_type, channel_size, bregma, site_face)
    print(region_label, region_length, region_channels)
    enter_coords = new_sp * vxsize_um
    end_coords = new_ep * vxsize_um

    dv = (new_sp[2] - new_ep[2]) * vxsize_um
    ap_tilt, ml_tilt = get_tilt_info(new_sp, new_ep)

    label_names, label_acronym, label_color, left_sites_color, right_sites_color = get_label_name(
        label_info, region_label, sites_label)

    # sites_color = [left_sites_color, right_sites_color]
    # print(label_names, label_acronym, label_color, chn_line_color)

    left_merged_labels, left_merged_colors, left_block_count = block_same_label_one_side(
        sites_label[:, 0], left_sites_color)
    right_merged_labels, right_merged_colors, right_block_count = block_same_label_one_side(
        sites_label[:, 1], right_sites_color)

    merged_labels = [left_merged_labels, right_merged_labels]
    merged_colors = [left_merged_colors, right_merged_colors]
    block_count = [left_block_count, right_block_count]

    # print(merged_labels, merged_colors, block_count)

    da_dict = {'object_name': 'probe', 'data': data, 'ap_tilt': ap_tilt, 'ml_tilt': ml_tilt,
               'insertion_coords_3d': start_pnt, 'terminus_coords_3d': end_pnt,
               'new_insertion_coords_3d': new_sp, 'new_terminus_coords_3d': new_ep,
               'direction': direction, 'probe_length': probe_length_with_tip, 'dv': dv,
               'ap_angle': ap_angle, 'ml_angle': ml_angle,
               'insertion_coords': enter_coords, 'insertion_vox': new_start_vox,
               'terminus_coords': end_coords, 'terminus_vox': new_end_vox,
               'sites_label': sites_label, 'sites_loc_related_to_Bregma': sites_loc,
               'merged_labels': merged_labels, 'merged_color': merged_colors, 'block_count': block_count,
               'region_label': region_label, 'region_length': region_length, 'region_channels': region_channels,
               'label_name': label_names, 'label_acronym': label_acronym, 'label_color': label_color}
    return da_dict


def get_region_label(data, label_data, bregma):
    region_label = []
    for i in range(len(data)):
        temp = data[i] + bregma
        temp = temp.astype(int)
        region_label.append(label_data[temp[0], temp[1], temp[2]])
    return region_label


def get_region_label_info(region_label, label_info):
    unique_label = np.unique(region_label)
    label_names = []
    label_acronym = []
    label_color = []
    region_count = []
    for i in range(len(unique_label)):
        # print(unique_label[i])
        if unique_label[i] == 0:
            label_names.append(' ')
            label_acronym.append(' ')
            label_color.append((128, 128, 128))
        else:
            da_ind = np.where(label_info['index'] == unique_label[i])[0][0]
            label_names.append(label_info['label'][da_ind])
            label_acronym.append(label_info['abbrev'][da_ind])
            label_color.append(label_info['color'][da_ind])

        region_count.append(len(np.where(np.ravel(region_label) == unique_label[i])[0]))

    return region_count, label_names, label_acronym, label_color


def calculate_virus_info(data, label_data, label_info, bregma):
    region_label = get_region_label(data, label_data, bregma)
    region_count, label_names, label_acronym, label_color = get_region_label_info(region_label, label_info)

    res_dict = {'object_name': 'virus', 'data': data, 'label_name': label_names,
                'label_acronym': label_acronym, 'label_color': label_color}

    return res_dict


def calculate_cells_info(data, label_data, label_info, bregma):
    region_label = get_region_label(data, label_data, bregma)
    region_count, label_names, label_acronym, label_color = get_region_label_info(region_label, label_info)

    res_dict = {'object_name': 'cell', 'data': data, 'label_name': label_names,
                'label_acronym': label_acronym, 'label_color': label_color, 'region_count': region_count}
    return res_dict


def order_contour_pnt(pnt):
    order_ind = []
    x_min = np.min(pnt[:, 0])
    left_ind = np.where(pnt[:, 0] == x_min)[0]
    if len(left_ind) > 1:
        low_ind = np.where(pnt[left_ind, :] == np.min(pnt[left_ind, :]))[0]
        left_ind = left_ind[low_ind]
    left_pnt = pnt[left_ind, :]
    lower_inds = np.where(pnt[:, 1] <= left_pnt[1])[0]
    lower_pnts = pnt[:, lower_inds]


def calculate_contour_line(data):
    data = np.asarray(data)
    res = splprep([data[:, 0], data[:, 1], data[:, 2]], s=2)
    tck = res[0]
    # x_knots, y_knots, z_knots = splev(tck[0], tck)
    u_fine = np.linspace(0, 1, len(data))
    x_fine, y_fine, z_fine = splev(u_fine, tck)
    pnts = np.stack([x_fine, y_fine, z_fine], axis=1)
    print(pnts)
    return pnts


def get_object_vis_color(color):
    vis_color = (color[0] / 255, color[1] / 255, color[2] / 255, 1)
    return vis_color


def create_plot_points_in_3d(data_dict):
    pnts = data_dict['data']
    vis_color = get_object_vis_color(data_dict['vis_color'])
    vis_points = gl.GLScatterPlotItem(pos=pnts, color=vis_color, size=3)
    return vis_points


def create_probe_line_in_3d(data_dict):
    pos = np.stack([data_dict['new_insertion_coords_3d'], data_dict['new_terminus_coords_3d']], axis=0)
    vis_color = get_object_vis_color(data_dict['vis_color'])
    probe_line = gl.GLLinePlotItem(pos=pos, color=vis_color, width=2, mode='line_strip')
    return probe_line


def create_drawing_line_in_3d(data_dict):
    pos = np.asarray(data_dict['data'])
    vis_color = get_object_vis_color(data_dict['vis_color'])
    if np.all(pos[0] == pos[-1]):
        drawing_obj = gl.GLSurfacePlotItem(x=pos[:, 0], y=pos[:, 1], z=pos[:, 2], colors=vis_color)
    else:
        drawing_obj = gl.GLLinePlotItem(pos=pos, color=vis_color, width=2, mode='line_strip')
    return drawing_obj


def create_contour_line_in_3d(data_dict):
    temp = data_dict['data'].tolist()
    if temp[0] != temp[-1]:
        temp.append(temp[0])
    pos = np.asarray(temp)
    vis_color = get_object_vis_color(data_dict['vis_color'])
    contour_obj = gl.GLLinePlotItem(pos=pos, color=vis_color, width=2, mode='line_strip')
    return contour_obj

    
def check_plotting_2d(points):
    points = np.asarray(points)


def hex2rgb(hex):
    if '#' in hex:
        hex = hex.lstrip('#')
        rgb_color = [int(hex[i:i + 2], 16) for i in (0, 2, 4)]
    else:
        if len(hex) == 6:
            rgb_color = [int(hex[i:i + 2], 16) for i in (0, 2, 4)]
    return rgb_color[0], rgb_color[1], rgb_color[2]


# @jit()
def hsv2rgb(h, s, v):
    # h [0, 1], s, v [0, 1]
    h = h * 360
    c = v * s
    m = v - c
    x = c * (1 - np.abs(np.mod(h / 60, 2) - 1))
    if 0 <= h < 60:
        r, g, b = (c + m, x + m, m)
    elif 60 <= h < 120:
        r, g, b = (x + m, c + m, m)
    elif 120 <= h < 180:
        r, g, b = (m, c + m, x + m)
    elif 180 <= h < 240:
        r, g, b = (m, x + m, c + m)
    elif 240 <= h < 300:
        r, g, b = (x + m, m, c + m)
    elif 300 <= h < 360:
        r, g, b = (c + m, m, x + m)
    else:
        r, g, b = (m, m, m)
    r = r * 65535
    b = b * 65535
    g = g * 65535
    return r, g, b


def color_img(h, s, v):
    r, g, b = hsv2rgb(h, s, v)
    da_image = cv2.merge((r, g, b))
    return da_image


def merge_channels_into_single_img(czi_img, channel_colors):
    merged_img = np.zeros((czi_img.shape[0], czi_img.shape[1], 3))
    frac = 1 / len(channel_colors)
    print(frac)
    for i in range(len(channel_colors)):
        temp_v = czi_img[:, :, i] / 65535
        temp_h = channel_colors[i][0]
        temp_s = channel_colors[i][1]
        print(temp_h, temp_s)
        da_img = color_img(temp_h, temp_s, temp_v)
        merged_img = merged_img + frac * da_img
    return merged_img


def make_color_lut(channel_color: tuple, bit_level: int):
    r, g, b = colorsys.hsv_to_rgb(channel_color[0], channel_color[1], channel_color[2])
    colors = [(0, 0, 0), (r * 255, g * 255, b * 255)]
    color_map = pg.ColorMap(pos=[0, 1], color=colors)
    da_lut = color_map.getLookupTable(nPts=bit_level, mode=pg.ColorMap.FLOAT)
    da_lut = da_lut * 255
    return da_lut


def get_qhsv_from_czi_hsv(hsv_color: tuple):
    """
    Get QColor HSV color format (360, 255, 255) from czi HSV color format (1, 1, 1).
    :param hsv_color: (float [0., 1.], float [0., 1.], float [0., 1.])
    :return: QColor
    """
    h_val = int(hsv_color[0] * 360)
    s_val = int(hsv_color[1] * 255)
    v_val = int(hsv_color[2] * 255)
    da_color = (h_val, s_val, v_val)
    return da_color


def gamma_line(input, lims, gamma, depth_level):
    inv_gamma = 1.0 / gamma
    y = np.zeros(len(input))
    inds = np.logical_and(input >= lims[0], input <= lims[1])
    y[inds] = np.power((input[inds] - lims[0]) / (lims[1] - lims[0]), gamma) * depth_level
    y[input <= lims[0]] = 0
    y[input >= lims[1]] = depth_level
    return y


def crop_landscape(image, dim):
    r = (dim[0] / image.shape[0]) / (dim[0] / dim[1])
    nw = int(image.shape[1] * r)

    resized = cv2.resize(image, (nw, int(dim[1])), interpolation=cv2.INTER_AREA)

    half_width = int(dim[0]) / 2
    half_shape_width = int(resized.shape[1]) / 2

    start_x = half_shape_width - half_width
    end_x = half_width + half_shape_width
    cropped = resized[0:dim[1], start_x:end_x]

    return cropped


def crop_portrait(image, dim):
    r = dim[1] / image.shape[1] / (dim[1] / dim[0])
    nh = int(image.shape[0] * r)

    resized = cv2.resize(image, (int(dim[0]), nh), interpolation=cv2.INTER_AREA)
    half_height = int(dim[1]) / 2
    half_shape_height = int(resized.shape[0]) / 2

    start_y = half_shape_height - half_height
    end_y = half_height + half_shape_height
    cropped = resized[start_y:end_y, 0:dim[0]]

    return cropped


def create_other_size(image, file_name, dim, location):
    # 1 => width index, 0 => height index
    if image.shape[0] > image.shape[1]:
      cropped = crop_portrait(image, dim)
    else:
      cropped = crop_landscape(image, dim)

    cv2.imwrite(os.path.join(location, file_name), cropped)


def make_hist_data(image_data, max_val):
    hist_data_list = []
    for i in range(image_data.shape[2]):
        if np.max(image_data[:, :, i]) == 0:
            da_bins = max_val
        else:
            da_bins = np.max(image_data[:, :, i])
        hist_y, x = np.histogram(image_data[:, :, i], bins=da_bins)
        y = np.log1p(hist_y)
        y = y / np.max(y) * max_val
        y = np.append(y, 0)
        sfunc = interp1d(x, y, 'cubic')
        inter_x = np.linspace(np.min(x), np.max(x), 200)
        inter_y = sfunc(inter_x)
        inter_y[inter_y < 0] = 0
        hist_data_list.append([inter_x, inter_y])
    return hist_data_list


def rect_contains(rect, point):
    if point[0] < rect[0]:
        return False
    elif point[1] < rect[1]:
        return False
    elif point[0] > rect[2] + rect[0]:
        return False
    elif point[1] > rect[3] + rect[1]:
        return False
    return True


def get_warp_matrix(src_tri_pnts, dst_tri_pnts):
    # Given a pair of triangles, find the affine transform.
    warp_mat = cv2.getAffineTransform(np.float32(src_tri_pnts), np.float32(dst_tri_pnts))
    return warp_mat


def apply_affine_transform(src_img, warp_mat, size_dst):
    # Apply the Affine Transform just found to the src image
    dst = cv2.warpAffine(src_img.astype(np.float32), warp_mat, (size_dst[0], size_dst[1]), None,
                         flags=cv2.INTER_LINEAR, borderMode=cv2.BORDER_REFLECT_101)
    return dst


def warp_triangle(img1, img2, t1, t2, is_rgb=False):

    # Find bounding rectangle for each triangle
    r1 = cv2.boundingRect(t1.astype(np.float32))
    r2 = cv2.boundingRect(t2.astype(np.float32))

    # Offset points by left top corner of the respective rectangles
    t1_rect = t1 - r1[:2]
    t2_rect = t2 - r2[:2]
    t2_rect_int = t2_rect.astype(int)

    # Get mask by filling triangle
    if is_rgb:
        mask = np.zeros((r2[3], r2[2], img1.shape[2]), dtype=np.float32)
        cv2.fillConvexPoly(mask, np.int32(t2_rect_int), tuple(np.repeat(1.0, img1.shape[2])), 16, 0)
    else:
        mask = np.zeros((r2[3], r2[2]), dtype=np.float32)
        cv2.fillConvexPoly(mask, np.int32(t2_rect_int), 1, 16, 0)

    # Apply warpImage to small rectangular patches
    img1_rect = img1[r1[1]:r1[1] + r1[3], r1[0]:r1[0] + r1[2]]
    # img2Rect = np.zeros((r2[3], r2[2]), dtype = img1Rect.dtype)

    size = (r2[2], r2[3])

    warp_mat = get_warp_matrix(t1_rect, t2_rect)
    img2_rect = apply_affine_transform(img1_rect, warp_mat, size)
    img2_rect = img2_rect * mask

    # Copy triangular region of the rectangular patch to the output image
    yr = (r2[0], r2[0] + r2[2])
    xr = (r2[1], r2[1] + r2[3])
    if is_rgb:
        img2[xr[0]:xr[1], yr[0]:yr[1]] = img2[xr[0]:xr[1], yr[0]:yr[1]] * (tuple(np.repeat(1.0, img1.shape[2])) - mask)
    else:
        img2[xr[0]:xr[1], yr[0]:yr[1]] = img2[xr[0]:xr[1], yr[0]:yr[1]] * (1 - mask)
    img2[xr[0]:xr[1], yr[0]:yr[1]] = img2[xr[0]:xr[1], yr[0]:yr[1]] + img2_rect


def warp_points(pnts, t1, t2):
    # Find bounding rectangle for each triangle
    # r1 = cv2.boundingRect(t1.astype(np.float32))
    # r2 = cv2.boundingRect(t2.astype(np.float32))
    # print(r1)
    # print(r2)

    # pnts =

    da_pnts = np.hstack([pnts, np.ones((len(pnts), 1))])
    # Offset points by left top corner of the respective rectangles
    # t1_rect = t1 - r1[:2]
    # t2_rect = t2 - r2[:2]

    warp_mat = get_warp_matrix(t1, t2)
    output = np.dot(warp_mat, da_pnts.T).T

    return output



# calculate delanauy triangle
def calculateDelaunayTriangles(rect, points):
    # create subdiv
    subdiv = cv2.Subdiv2D(rect)

    # Insert points into subdiv
    for p in points:
        subdiv.insert(p)

    triangleList = subdiv.getTriangleList()

    delaunayTri = []

    pt = []

    for t in triangleList:
        pt.append((t[0], t[1]))
        pt.append((t[2], t[3]))
        pt.append((t[4], t[5]))

        pt1 = (t[0], t[1])
        pt2 = (t[2], t[3])
        pt3 = (t[4], t[5])

        if rect_contains(rect, pt1) and rect_contains(rect, pt2) and rect_contains(rect, pt3):
            ind = []
            for j in range(0, 3):
                for k in range(0, len(points)):
                    if (abs(pt[j][0] - points[k][0]) < 1.0 and abs(pt[j][1] - points[k][1]) < 1.0):
                        ind.append(k)
            if len(ind) == 3:
                delaunayTri.append((ind[0], ind[1], ind[2]))

        pt = []

    return delaunayTri


def get_vertex_ind_in_triangle(subdiv):
    triangles = subdiv.getTriangleList()
    n_triangles = len(triangles)
    tri_vet_inds = []
    for i in range(n_triangles):
        da_triangle = triangles[i]
        p1 = [da_triangle[0], da_triangle[1]]
        p2 = [da_triangle[2], da_triangle[3]]
        p3 = [da_triangle[4], da_triangle[5]]
        tri_vet_inds.append([subdiv.locate(p1)[2], subdiv.locate(p2)[2], subdiv.locate(p3)[2]])
    tri_vet_inds = np.asarray(tri_vet_inds) - 4
    return tri_vet_inds


def get_pnts_triangle_ind(tri_vet_inds, tri_data, size, pnts):
    # import cv2
    # import numpy as np
    # img_rec = (0, 0, 100, 200)
    # da_triangle = np.array([[0, 0], [0, 50], [100, 100]])
    # size = (200, 100)
    # pnts = np.array([[0, 1], [0, 2], [60, 200]])

    update_pnts = pnts.copy()
    n_pnts = len(pnts)
    loc = np.zeros(n_pnts)
    loc[:] = np.nan
    da_order = []

    ct_list = []
    for i in range(len(tri_vet_inds)):
        da_inds = tri_vet_inds[i]
        da_triangle = np.array([tri_data[da_inds[0]], tri_data[da_inds[1]], tri_data[da_inds[2]]])
        mask = np.zeros(size, dtype=np.uint8)
        cv2.fillPoly(mask, pts=[da_triangle], color=255)
        ct, hc = cv2.findContours(image=mask, mode=cv2.RETR_TREE, method=cv2.CHAIN_APPROX_NONE)
        ct_list.append(ct[0])

        # range_y = (np.min(da_triangle[:, 1]), np.max(da_triangle[:, 1]))
        # range_x = (np.min(da_triangle[:, 0]), np.max(da_triangle[:, 0]))
        #
        # valid_pnts_ind = [ind for ind in range(n_pnts) if range_x[0] <= pnts[ind][0] <= range_x[1] and range_y[0] <= pnts[ind][1] <= range_y[1]]
        # valid_pnts_ind = [ind for ind in valid_pnts_ind if ind not in da_order]
        # valid_pnts = pnts[valid_pnts_ind]
        # for j in range(len(valid_pnts)):
        #     res = cv2.pointPolygonTest(ct_list[i], (int(pnts[j][0]), int(pnts[j][1])), True)
        #     if res >= 0:
        #         loc.append(i)
        #         da_order.append(valid_pnts_ind[j])

    #     temp = np.zeros(len(update_pnts))
    #
    #
    # for i in range(len(tri_vet_inds)):
    #     da_
    #
    for i in range(len(pnts)):
        for j in range(len(ct_list)):
            da_ct = ct_list[j]
            res = cv2.pointPolygonTest(da_ct, (int(pnts[i][0]), int(pnts[i][1])), False)
            if res >= 0:
                loc[i] = j
                break


    return loc




def get_sides_points(img_size):
    size0 = img_size[1] - 1
    size1 = img_size[0] - 1
    side_lines = np.asarray([[[0, 0], [size0, 0]], [[size0, 0], [size0, size1]],
                             [[0, size1], [size0, size1]], [[0, 0], [0, size1]]])
    corner_points = [[0, 0], [size0, 0], [size0, size1], [0, size1]]
    return side_lines, corner_points


def num_side_pnt_changed(num_pnt, corner_points, side_lines):
    n_pnts_to_make = num_pnt - 2
    onside_data = corner_points.copy()
    if n_pnts_to_make > 0:
        for i in range(4):
            da_line = side_lines[i]
            for j in range(n_pnts_to_make):
                inline_point = da_line[0] + (da_line[1] - da_line[0]) / (n_pnts_to_make + 1) * (j + 1)
                onside_data.append([int(inline_point[0]), int(inline_point[1])])
    return onside_data


def match_sides_points(rect_atlas, size_atlas, rect_image, size_image):
    x_factor = rect_atlas[2] / rect_image[2]
    y_factor = rect_atlas[3] / rect_image[3]

    actual_left_dist_image = size_image[1] - (rect_image[0] + rect_image[2])
    actual_bottom_dist_image = size_image[0] - (rect_image[1] + rect_image[3])

    actual_left_dist_atlas = size_atlas[1] - (rect_atlas[0] + rect_atlas[2])
    actual_bottom_dist_atlas = size_atlas[0] - (rect_atlas[1] + rect_atlas[3])

    left_dist_atlas = rect_image[0] * x_factor
    right_dist_atlas = actual_left_dist_image * x_factor

    top_dist_atlas = rect_image[1] * y_factor
    bottom_dist_atlas = actual_bottom_dist_image * y_factor

    if left_dist_atlas <= rect_atlas[0]:
        atlas_corner_x = int(rect_atlas[0] - left_dist_atlas)
        image_corner_x = 0
    else:
        atlas_corner_x = 0
        left_dist_image = rect_atlas[0] / x_factor
        image_corner_x = int(rect_image[0] - left_dist_image)

    if right_dist_atlas <= actual_left_dist_atlas:
        atlas_width = int(size_atlas[1] - atlas_corner_x - (actual_left_dist_atlas - right_dist_atlas))
        image_width = size_image[1]
    else:
        atlas_width = size_atlas[1] - atlas_corner_x
        right_dist_image = actual_left_dist_atlas / x_factor
        image_width = int(size_image[1] - image_corner_x - (actual_left_dist_image - right_dist_image))

    if top_dist_atlas <= rect_atlas[1]:
        atlas_corner_y = int(rect_atlas[1] - top_dist_atlas)
        image_corner_y = 0
    else:
        atlas_corner_y = 0
        top_dist_image = rect_atlas[1] / y_factor
        image_corner_y = int(rect_image[1] - top_dist_image)

    if bottom_dist_atlas <= actual_bottom_dist_atlas:
        atlas_height = int(size_atlas[0] - atlas_corner_y - (actual_bottom_dist_atlas - bottom_dist_atlas))
        image_height = size_image[0]
    else:
        atlas_height = size_atlas[0] - atlas_corner_y
        bottom_dist_image = actual_bottom_dist_atlas / y_factor
        image_height = int(size_image[0] - image_corner_y - (actual_bottom_dist_image - bottom_dist_image))

    atlas_rect = (atlas_corner_x, atlas_corner_y, atlas_width, atlas_height)
    image_rect = (image_corner_x, image_corner_y, image_width, image_height)

    atlas_corners, atlas_lines = get_corner_line_from_rect(atlas_rect)
    image_corners, image_lines = get_corner_line_from_rect(image_rect)

    return atlas_corners, atlas_lines, image_corners, image_lines


def get_corner_line_from_rect(rect):
    corners = [[rect[0], rect[1]], [rect[0] + rect[2] - 1, rect[1]],
               [rect[0] + rect[2] - 1, rect[1] + rect[3] - 1], [rect[0], rect[1] + rect[3] - 1]]

    lines = np.asarray([[corners[0], corners[1]], [corners[1], corners[2]],
                        [corners[3], corners[2]], [corners[0], corners[3]]])

    return corners, lines


def make_label_rgb_img(label_img, lut):
    fimg = np.dstack([label_img, label_img, label_img])
    unique_label = np.unique(label_img).astype(int)
    for ind in np.unique(unique_label):
        if ind == 0:
            continue
        loc = np.where(label_img == ind)
        print(lut[ind])
        fimg[loc[0], loc[1], 0] = lut[ind][0]
        fimg[loc[0], loc[1], 1] = lut[ind][1]
        fimg[loc[0], loc[1], 2] = lut[ind][2]
    return fimg.astype('uint8')


def make_contour_img(lable_img):
    unique_label = np.unique(lable_img).astype('int')
    img_size = lable_img.shape
    contour_img = np.zeros(img_size, 'uint8')
    for label_ind in unique_label:
        if label_ind == 0:
            continue
        temp = np.zeros(img_size, 'uint8')
        temp[lable_img == label_ind] = 1
        ct, hc = cv2.findContours(image=temp, mode=cv2.RETR_TREE, method=cv2.CHAIN_APPROX_NONE)
        for j in range(len(ct)):
            da_contour = ct[j].copy()
            da_shp = da_contour.shape
            da_contour = np.reshape(da_contour, (da_shp[0], da_shp[2]))
            contour_img[da_contour[:, 1], da_contour[:, 0]] = 1
    return contour_img


def get_tri_lines(rect, pnts):
    subdiv = cv2.Subdiv2D(rect)
    for p in pnts:
        subdiv.insert((int(p[0]), int(p[1])))
    edge_list = subdiv.getEdgeList()
    lines_list = []
    # special_pnt = []
    for el in edge_list:
        pt1 = [el[0], el[1]]
        pt2 = [el[2], el[3]]
        if rect_contains(rect, pt1) and rect_contains(rect, pt2):
            lines_list.append(el)
        # else:
        #     if not rect_contains(rect, pt1) and pt1 not in special_pnt:
        #         special_pnt.append(pt1)
        #     if not rect_contains(rect, pt2) and pt2 not in special_pnt:
        #         special_pnt.append(pt2)
    return lines_list  #, special_pnt


# czi_img = image_file.data['scene 0'].copy()
# channel_hsv = image_file.hsv_colors
# temp_img = merge_channels_into_single_img(czi_img, channel_hsv)

def gamma_correction(src, gamma):
    inv_gamma = 1 / gamma

    table = [((i / 255) ** inv_gamma) * 255 for i in range(256)]
    table = np.array(table, np.uint8)

    return cv2.LUT(src, table)



def get_lower_val(val, tol, lim):
    lower_val = val - tol if tol < val else lim
    return lower_val


def get_upper_val(val, tol, lim):
    upper_val = val + tol
    upper_val = upper_val if upper_val <= lim else lim
    return upper_val


def get_bound_color(color, tol, level, mode):
    tol = float(tol)
    if mode == 'gray':
        lower_val = get_lower_val(color, tol, 0)
        upper_val = get_upper_val(color, tol, level)
    elif mode == 'hsv':
        lower_val = [get_lower_val(color[0], 1., 0)]
        upper_val = [get_upper_val(color[0], 1., 180)]
        for i in range(1, 3):
            lower_val.append(get_lower_val(color[i], tol, 0))
            upper_val.append(get_upper_val(color[i], tol, 255))
    else:
        lower_val = []
        upper_val = []
        for i in range(3):
            lower_val.append(get_lower_val(color[i], tol, 0))
            upper_val.append(get_upper_val(color[i], tol, level))
    return lower_val, upper_val


# ----------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------
def render_volume(atlas_data, atlas_folder, factor=2, level=0.1):
    da_data = atlas_data.copy()
    img = np.ascontiguousarray(da_data[::factor, ::factor, ::factor])
    verts, faces = pg.isosurface(ndi.gaussian_filter(img.astype('float64'), (2, 2, 2)), np.max(da_data) * level)

    md = gl.MeshData(vertexes=verts * factor, faces=faces)

    outfile = open(os.path.join(atlas_folder, 'atlas_meshdata.pkl'), 'wb')
    pickle.dump(md, outfile)
    outfile.close()

    return md


def render_small_volume(label_id, save_path, atlas_data, atlas_label, factor=2, level=0.1):
    temp_atlas = atlas_data.copy()
    temp_atlas[atlas_label != label_id] = 0
    pimg = np.ascontiguousarray(temp_atlas[::factor, ::factor, ::factor])
    verts, faces = pg.isosurface(ndi.gaussian_filter(pimg.astype('float64'), (2, 2, 2)), np.max(temp_atlas) * level)
    # small_verts_list[str(id)] = verts
    # small_faces_list[str(id)] = faces

    md = gl.MeshData(vertexes=verts * factor, faces=faces)

    outfile = open(os.path.join(save_path, '{}.pkl'.format(label_id)), 'wb')
    pickle.dump(md, outfile)
    outfile.close()


def get_statusbar_style(col):
    style = 'background-color: #323232;' \
            ' color: {}; border-top: 1px solid #272727; ' \
            'padding-left: 30px;'.format(col)
    return style


def rotate(image, angle, img_center=None, scale=1.0):
    (h, w) = image.shape[:2]

    if img_center is None:
        img_center = (w // 2, h // 2)

    rot_mat = cv2.getRotationMatrix2D(img_center, angle, scale)
    rotated_img = cv2.warpAffine(image, rot_mat, (w, h))

    return rotated_img


def rotate_bound(image, angle):
    (h, w) = image.shape[:2]
    (center_x, center_y) = (w / 2, h / 2)

    rot_mat = cv2.getRotationMatrix2D((center_x, center_y), -angle, 1.0)
    cos = np.abs(rot_mat[0, 0])
    sin = np.abs(rot_mat[0, 1])

    bound_w = int((h * sin) + (w * cos))
    bound_h = int((h * cos) + (w * sin))

    rot_mat[0, 2] += (bound_w / 2) - center_x
    rot_mat[1, 2] += (bound_h / 2) - center_y

    return cv2.warpAffine(image, rot_mat, (bound_w, bound_h))


def center_resize(img, dim):
    img_shape = img.shape
    width, height = img_shape[1], img_shape[0]
    scale_factor = np.min(np.array([dim[0] / width, dim[1] / height]))
    resize_dim = (int(width * scale_factor), int(height * scale_factor))
    resize_img = cv2.resize(img, resize_dim, interpolation=cv2.INTER_LINEAR)

    y = int(0.5 * (dim[0] - resize_dim[0]))
    x = int(0.5 * (dim[1] - resize_dim[1]))

    if len(img_shape) == 3:
        center_img = np.zeros((dim[1], dim[0], img.shape[2])).astype(img.dtype)
    else:
        center_img = np.zeros((dim[1], dim[0])).astype(img.dtype)

    center_img[x:(x + resize_dim[1]), y:(y + resize_dim[0])] = resize_img

    return center_img


def get_tb_size(img_size):
    scale_factor = np.max(np.ravel(img_size) / 80)
    tb_size = (int(img_size[1] / scale_factor), int(img_size[0] / scale_factor))
    return tb_size


def get_slice_atlas_coord(points, cut, size, width, height, distance, origin):
    """

    :param points: pnts on image
    :param cut: the cut of slice image
    :param size: the size (height, width) of slice image in pixel
    :param width: width of image in mm
    :param height: height of image in mm
    :param distance: distance of slice with respect to Bregma in mm
    :param origin: the coord of bregma on the current slice image
    :return:
    """
    width_factor = width / size[1] * 1000
    height_factor = height / size[0] * 1000
    if cut == 'Coronal':
        y_val = np.repeat(distance * 1000, len(points))
        x_val = (points[:, 0] - origin[0]) * width_factor
        z_val = (points[:, 1] - origin[1]) * height_factor
    elif cut == 'Sagittal':
        x_val = np.repeat(distance * 1000, len(points))
        y_val = (points[:, 0] - origin[0]) * width_factor
        z_val = (points[:, 1] - origin[1]) * height_factor
    else:
        z_val = np.repeat(distance * 1000, len(points))
        y_val = (points[:, 0] - origin[0]) * width_factor
        x_val = (points[:, 1] - origin[1]) * height_factor
    return x_val, y_val, z_val


def delete_points_inside_eraser(points, ct, r):
    x_min, y_min, x_max, y_max = ct[0] - r, ct[1] - r, ct[0] + r, ct[1] + r
    x_bool = np.logical_and(points[:, 0] <= x_max, points[:, 0] >= x_min)
    y_bool = np.logical_and(points[:, 1] <= y_max, points[:, 1] >= y_min)
    chk_inds = np.where(np.logical_and(x_bool, y_bool))[0]
    if len(chk_inds) == 0:
        return None, None
    chk_pnt = points[chk_inds]
    dist = np.sum(np.power(chk_pnt - ct, 2), 1)
    del_ind = np.where(dist < np.power(r, 2))[0]
    if len(del_ind) == 0:
        return None, None
    real_del_ind = chk_inds[del_ind]
    remain_inds = np.zeros(len(points)) < 1
    remain_inds[real_del_ind] = False
    remain_points = points[remain_inds]
    return remain_points, real_del_ind


def interpolate_contour_points(points):
    if not np.all(points[-1] == points[0]):
        points[:, 0] = np.r_[points[:, 0], points[:, 0]]
        points[:, 1] = np.r_[points[:, 1], points[:, 1]]

    tck = splprep([points[:, 0], points[:, 1]], s=0, per=True)

    xi, yi = splev(np.linspace(0, 1, 1000), tck[0])
    return xi, yi


def create_vis_img(size, point_data, color, vis_type='p', closed=False):
    if isinstance(point_data, list):
        point_data = np.asarray(point_data).astype(int)
    img = np.zeros((size[0], size[1], 3), 'uint8')
    da_color = (int(color[0]), int(color[1]), int(color[2]))
    if vis_type == 'p':
        for i in range(len(point_data)):
            cv2.circle(img, (int(point_data[i][0]), int(point_data[i][1])), radius=5, color=da_color, thickness=-1)
    else:
        if closed:
            cv2.fillPoly(img, pts=[point_data], color=da_color)
        else:
            for i in range(len(point_data) - 1):
                cv2.line(img, (int(point_data[i][0]), int(point_data[i][1])),
                         (int(point_data[i + 1][0]), int(point_data[i + 1][1])),
                         color=da_color, thickness=8)
    return img


def color_vis_img(img, color):
    temp = np.zeros((img.shape[0], img.shape[1], 3), 'uint8')
    temp[img != 0, 0] = color[0]
    temp[img != 0, 1] = color[1]
    temp[img != 0, 2] = color[2]
    return temp


def check_loaded_project(project_dict):
    all_keys = list(project_dict.keys())
    valid_keys = ['atlas_path', 'current_atlas', 'num_windows', 'probe_type', 'np_onside', 'processing_slice',
                  'processing_img', 'overlay_img', 'atlas_control', 'img_ctrl_data', 'setting_data', 'tool_data',
                  'layer_data', 'working_img_data', 'working_atlas_data', 'object_data']
    valid_project = True
    for da_key in all_keys:
        if da_key not in valid_keys:
            valid_project = False
    return valid_project


def check_bounding_contains(points, size):
    if np.any(points[:, 0] > size[1]) or np.any(points[:, 1] > size[0]):
        return False
    else:
        return True


def obj_data_to_mesh3d(filename):
    vertices = []
    faces = []

    with open(filename, 'r') as objf:
        for line in objf:
            slist = line.split()
            if slist:
                if slist[0] == 'v':
                    vertex = np.array(slist[1:], dtype=float)
                    vertices.append(vertex)
                elif slist[0] == 'f':
                    face = []
                    for k in range(1, len(slist)):
                        face.append([int(s) for s in slist[k].replace('//', '/').split('/')])
                    if len(face) > 3:  # triangulate the n-polyonal face, n>3
                        faces.extend(
                            [[face[0][0] - 1, face[k][0] - 1, face[k + 1][0] - 1] for k in range(1, len(face) - 1)])
                    else:
                        faces.append([face[j][0] - 1 for j in range(len(face))])
                else:
                    pass

    return np.array(vertices), np.array(faces)


def make_atlas_label_contour(atlas_folder, segmentation_data):
    sagital_contour_img = np.zeros(segmentation_data.shape, 'i')
    coronal_contour_img = np.zeros(segmentation_data.shape, 'i')
    horizontal_contour_img = np.zeros(segmentation_data.shape, 'i')

    # pre-process boundary
    for i in range(len(segmentation_data)):
        da_slice = segmentation_data[i, :, :].copy()
        contour_img = make_contour_img(da_slice)
        sagital_contour_img[i, :, :] = contour_img

    for i in range(segmentation_data.shape[1]):
        da_slice = segmentation_data[:, i, :].copy()
        contour_img = make_contour_img(da_slice)
        coronal_contour_img[:, i, :] = contour_img

    for i in range(segmentation_data.shape[2]):
        da_slice = segmentation_data[:, :, i].copy()
        contour_img = make_contour_img(da_slice)
        horizontal_contour_img[:, :, i] = contour_img

    boundary = {'s_contour': sagital_contour_img,
                'c_contour': coronal_contour_img,
                'h_contour': horizontal_contour_img}

    bnd = {'data': boundary}

    outfile_ct = open(os.path.join(atlas_folder, 'contour_pre_made.pkl'), 'wb')
    pickle.dump(bnd, outfile_ct)
    outfile_ct.close()

    return boundary


def get_angle_two_vector(vec1, vec2):
    """
    in 2d
    :param vec1:
    :param vec2:
    :return:
    """

    vec = np.array([-vec2[1], vec2[0]])
    b_coord = np.dot(vec1, vec2)
    p_coord = np.dot(vec1, vec)
    da_ang = np.arctan2(p_coord, b_coord)

    return da_ang


def rotate_base_points(data, base_loc):
    temp = np.stack([base_loc, np.array([0, 0, 0, 0])], axis=1)
    vec1 = data[1] - data[0]
    vec2 = np.array([0, 1])
    ang = get_angle_two_vector(vec1, vec2)
    rotm = np.array([[np.cos(ang), -np.sin(ang)], [np.sin(ang), np.cos(ang)]])
    base_pnt = np.dot(rotm, temp.T).T
    start_pnt = base_pnt + data[0]
    end_pnt = base_pnt + data[1]
    return start_pnt, end_pnt


import os
import sys
import numpy as np
from PyQt5.QtWidgets import *
from PyQt5.QtGui import *
from PyQt5.QtCore import *
from pyqtgraph.Qt import QtGui, QtCore
from .uuuuuu import hsv2rgb, gamma_line, get_qhsv_from_czi_hsv, make_hist_data
from .styles import Styles

channel_button_style = '''
QPushButton {
    margin: 0 px; 
    padding: 0 px; 
    border-color: transparent; 
    background-color: #747a80;
    color: white;
    border-top-right-radius: 5px;
    border-top-left-radius: 5px;
    border-bottom-right-radius: 0px;
    border-bottom-left-radius: 0px;
} 

QLabel {
    border-radius: 0px;
    border-top: 5px solid rgb(255, 255, 255);
    border-bottom: None;
    border-left: None;
    border-right: None;
}

'''


color_combo_style = '''
/*---------------------- QComboBox -----------------------*/
QComboBox {
    border-left: 1px solid gray;
    border-right: 1px solid gray;
    border-bottom: 1px solid gray;
    border-top: None;
    border-top-right-radius: 0px;
    border-top-left-radius: 0px;
    border-bottom-right-radius: 5px;
    border-bottom-left-radius: 5px;
    padding: 0px 3px 0px 3px;
    background-color: transparent;
    color: white;
    margin: 0px;
}

QComboBox:item {
    background: #323232;
    color: white;
    min-height: 10px;
    margin: 0px;
}

QComboBox:item:selected
{
    border: None;
    background: #232323;
    margin: 0px;
}

QComboBox:editable {
    background: transparent;
}

QComboBox:!editable, QComboBox::drop-down:editable {
    background: transparent;
    border-left: 1px solid gray;
    border-right: 1px solid gray;
    border-bottom: 1px solid gray;
    border-top: None;
    border-top-right-radius: 0px;
    border-top-left-radius: 0px;
    border-bottom-right-radius: 5px;
    border-bottom-left-radius: 5px;
}



/* QComboBox gets the "on" state when the popup is open */
QComboBox:!editable:on, QComboBox::drop-down:editable:on {
    background: transparent;
}

QComboBox:on { /* shift the text when the popup opens */
    padding: 3px;
    color: white;
    background-color: transparent;
    selection-background-color: transparent;
}

QComboBox::drop-down {
    subcontrol-origin: padding;
    subcontrol-position: top right;
    border-top: None;
    border-bottom: None;
    border-left-width: 1px;
    border-left-color: transparent;
    border-left-style: solid; /* just a single line */
    border-top-right-radius: 3px; /* same radius as the QComboBox */
    border-bottom-right-radius: 3px;
}

QComboBox::down-arrow {
    image: url(icons/tdown.svg);
    width: 13px;
    height: 14px;
    padding-right: 3px;
}

QComboBox::down-arrow:on { /* shift the arrow when popup is open */
    top: 1px;
    left: 1px;
}

'''

text_combo_list_style = '''
    QListView {
        background: #656565;
        border: 1px solid gray;
        color: white;
        border-radius: 0px;
    }

    QListView::item {
        border: None;
        background: transparent;
        margin:3px;
        height: 20px;
    }                             

    QListView::item:selected { 
        border: None;
        margin:3px;
        color: white;
        background: #232323; 
        height: 20px;
    }

    QListView::item:selected:!active {
        background: #323232;
        border: None;
    }

    QListView::item:selected:active {
        background: #323232;
        border: None;
    }

    QListView::item:hover {
        background: #323232;
        border: None;
    }
'''


class BWSpin(QWidget):
    def __init__(self, parent=None):
        QWidget.__init__(self)
        # wrap_frame = QFrame()
        wrap_layout = QVBoxLayout(self)
        wrap_layout.setContentsMargins(0, 0, 0, 0)
        wrap_layout.setSpacing(0)

        self.spin_nam = QLabel()
        self.spin_nam.setAlignment(Qt.AlignCenter)
        self.spin_val = QSpinBox()

        wrap_layout.addWidget(self.spin_nam)
        wrap_layout.addWidget(self.spin_val)


class GammaSpin(QWidget):
    def __init__(self, parent=None):
        QWidget.__init__(self)
        # wrap_frame = QFrame()
        wrap_layout = QVBoxLayout(self)
        wrap_layout.setContentsMargins(0, 0, 0, 0)
        wrap_layout.setSpacing(0)

        self.spin_nam = QLabel()
        self.spin_nam.setAlignment(Qt.AlignCenter)
        self.spin_val = QDoubleSpinBox()
        self.spin_val.setDecimals(2)

        wrap_layout.addWidget(self.spin_nam)
        wrap_layout.addWidget(self.spin_val)


class ColorCombo(QComboBox):
    def __init__(self, parent=None):
        QComboBox.__init__(self)
        # styles = Styles()
        self.setStyleSheet(color_combo_style)
        self.px = QPixmap(80, 30)
        n_vals = ['Light Blue', 'Blue', 'Green', 'Lime Green', 'Yellow', 'Olive', 'Red', 'Dark Red', 'Violet', 'Purple',
                  'Orange', 'Dark Orange', 'Turquoise', 'Blue Green']
        h_vals = [199, 235, 103, 115, 60, 60, 5, 4, 295, 279, 21, 20, 181, 151]
        s_vals = [60, 100, 70, 57, 67, 55, 85, 88, 77,  84, 81, 82, 51, 46]
        v_vals = [98, 96, 98, 59, 100, 60, 90, 54, 97,  77, 92, 73, 99, 79]
        s_vals = np.ravel(s_vals) * 0.01 * 255
        s_vals = s_vals.astype('uint8')
        v_vals = np.ravel(v_vals) * 0.01 * 255
        v_vals = v_vals.astype('uint8')
        self.hsv_color_list = []
        for i in range(len(h_vals)):
            self.hsv_color_list.append((h_vals[i] / 360., s_vals[i] / 255., v_vals[i] / 255.))
            self.px.fill(QColor.fromHsv(h_vals[i], s_vals[i], v_vals[i]))
            self.addItem(QIcon(self.px), n_vals[i])

        combo_list = QListView(self)
        combo_list.setStyleSheet(text_combo_list_style)
        self.setView(combo_list)


class ChannelSelector(QWidget):
    class SignalProxy(QtCore.QObject):
        visChannels = QtCore.Signal(object, object)
        changeColors = QtCore.Signal(object, object)

    def __init__(self, rgb=False, parent=None):
        self._sigprox = ChannelSelector.SignalProxy()
        self.sig_vis_channels = self._sigprox.visChannels
        self.sig_change_color = self._sigprox.changeColors

        QWidget.__init__(self)

        self.vis = True
        self.setFixedSize(60, 60)
        self.setStyleSheet(channel_button_style)

        # wrap_frame = QFrame()
        wrap_layout = QVBoxLayout(self)
        wrap_layout.setContentsMargins(0, 0, 0, 0)
        wrap_layout.setSpacing(0)

        self.vis_btn = QPushButton()
        self.vis_btn.setFixedSize(60, 30)
        self.vis_btn.setCheckable(True)
        self.vis_btn.clicked.connect(self.change_vis)
        self.color_combo = ColorCombo()
        self.color_combo.setFixedSize(60, 28)
        self.color_combo.currentIndexChanged.connect(self.selection_change)
        self.color_label = QLabel()

        self.color_label.setVisible(False)

        wrap_layout.addWidget(self.vis_btn)
        wrap_layout.addWidget(self.color_combo)
        wrap_layout.addWidget(self.color_label)

    def set_channel_index(self, index):
        self.index = index

    def change_vis(self):
        if self.vis_btn.isChecked():
            self.set_checked(True)
        else:
            self.set_checked(False)
        self.sig_vis_channels.emit(self.vis, self.index)

    def set_checked(self, is_checked):
        if is_checked:
            self.color_label.setVisible(True)
            self.color_combo.setVisible(False)
            self.vis = False
        else:
            self.color_label.setVisible(False)
            self.color_combo.setVisible(True)
            self.vis = True

    def selection_change(self):
        da_index = self.color_combo.currentIndex()
        da_color = self.color_combo.hsv_color_list[da_index]
        self.sig_change_color.emit(da_color, self.index)

    def add_item(self, hsv_color):
        self.color_combo.hsv_color_list.append((hsv_color[0], hsv_color[1], hsv_color[2]))
        q_hsv = get_qhsv_from_czi_hsv(hsv_color)
        self.color_combo.px.fill(QColor.fromHsv(q_hsv[0], q_hsv[1], q_hsv[2]))
        self.color_combo.addItem(QIcon(self.color_combo.px), '')
        self.color_combo.setCurrentIndex(len(self.color_combo.hsv_color_list) - 1)

    def delete_item(self):
        self.color_combo.hsv_color_list.pop()
        self.color_combo.removeItem(len(self.color_combo.hsv_color_list)-1)



from PyQt5.QtWidgets import *
from PyQt5.QtCore import *

dialog_style = '''
QDialog {
    background-color: rgb(50, 50, 50);
    color: white;
    font-size: 12px;
    width: 50px;
}

QLabel {
    color: white;
    width: 50px;
}
'''


class LayerSettingDialog(QDialog):
    def __init__(self, window_name, min_val, max_val, val):
        super().__init__()

        self.setWindowTitle(window_name)
        self.setStyleSheet(dialog_style)

        self.val = val

        self.val_slider = QSlider(Qt.Horizontal)
        self.val_slider.setValue(val)
        self.val_slider.setMinimum(min_val)
        self.val_slider.setMaximum(max_val)
        self.val_slider.setSingleStep(1)
        self.val_slider.sliderMoved.connect(self.val_spinbox_changed)

        self.val_spinbox = QSpinBox()
        self.val_spinbox.setValue(val)
        self.val_spinbox.setMaximum(max_val)
        self.val_spinbox.setMinimum(min_val)
        self.val_spinbox.setSingleStep(1)
        self.val_spinbox.valueChanged.connect(self.value_changed)

        # ok button, used to close window
        ok_btn = QDialogButtonBox(QDialogButtonBox.Ok)
        ok_btn.accepted.connect(self.accept)

        # add widget to layout
        layout = QHBoxLayout()
        layout.addWidget(self.val_slider)
        layout.addWidget(self.val_spinbox)
        layout.addSpacing(10)
        layout.addWidget(ok_btn)
        self.setLayout(layout)

    def accept(self) -> None:
        self.close()

    def val_spinbox_changed(self):
        val = self.val_slider.value()
        self.val = val
        self.val_spinbox.setValue(val)

    def value_changed(self):
        val = self.val_spinbox.value()
        self.val = val
        self.val_slider.setValue(val)


class SliceSettingDialog(QDialog):
    def __init__(self):
        super().__init__()

        self.setWindowTitle('Current Slice Settings')
        self.setStyleSheet(dialog_style)

        self.cut = 'Coronal'
        self.width = 0
        self.height = 0
        self.distance = 0

        self.cut_combo = QComboBox()
        self.cut_combo.addItems(['Coronal', 'Sagittal', 'Horizontal'])
        self.cut_combo.currentIndexChanged.connect(self.cut_changed)

        width_label = QLabel('Width (mm):')
        height_label = QLabel('Height (mm):')
        distance_label = QLabel('Distance w.r.t. Bregma (mm):')

        self.width_val = QDoubleSpinBox()
        self.width_val.setValue(0)
        self.width_val.setRange(-50, 50)
        self.width_val.valueChanged.connect(self.width_val_changed)
        self.height_val = QDoubleSpinBox()
        self.height_val.setValue(0)
        self.height_val.setRange(-50, 50)
        self.height_val.valueChanged.connect(self.height_val_changed)
        self.distance_val = QDoubleSpinBox()
        self.distance_val.setValue(0)
        self.distance_val.setRange(-50, 50)
        self.distance_val.valueChanged.connect(self.distance_val_changed)

        # ok button, used to close window
        ok_btn = QDialogButtonBox(QDialogButtonBox.Ok)
        ok_btn.accepted.connect(self.accept)

        # add widget to layout
        content_frame = QFrame()
        content_layout = QGridLayout(content_frame)
        content_layout.addWidget(self.cut_combo, 0, 0, 1, 2)
        content_layout.addWidget(width_label, 1, 0, 1, 1)
        content_layout.addWidget(self.width_val, 1, 1, 1, 1)
        content_layout.addWidget(height_label, 2, 0, 1, 1)
        content_layout.addWidget(self.height_val, 2, 1, 1, 1)
        content_layout.addWidget(distance_label, 3, 0, 1, 1)
        content_layout.addWidget(self.distance_val, 3, 1, 1, 1)

        layout = QVBoxLayout()
        layout.addWidget(content_frame)
        layout.addWidget(ok_btn)
        self.setLayout(layout)

    def accept(self) -> None:
        self.close()

    def cut_changed(self):
        self.cut = self.cut_combo.currentText()

    def width_val_changed(self):
        self.width = self.width_val.value()

    def height_val_changed(self):
        self.height = self.height_val.value()

    def distance_val_changed(self):
        self.distance = self.distance_val.value()


class QDoubleButton(QPushButton):
    right_clicked = pyqtSignal()
    left_clicked = pyqtSignal()
    double_clicked = pyqtSignal()

    def __init__(self, *args, **kwargs):
        super(QDoubleButton, self).__init__(*args, **kwargs)

        self.timer = QTimer()
        self.timer.setSingleShot(True)
        self.timer.setInterval(150)
        self.timer.timeout.connect(self.timeout)

        self.is_double = False
        self.is_left_click = True

        self.installEventFilter(self)

    def eventFilter(self, obj, event):
        if event.type() == QEvent.MouseButtonPress:
            if not self.timer.isActive():
                self.timer.start()

            self.is_left_click = False
            if event.button() == Qt.LeftButton:
                self.is_left_click = True

            return True

        elif event.type() == QEvent.MouseButtonDblClick:
            self.is_double = True
            return True

        return False

    def timeout(self):
        if self.is_double:
            self.double_clicked.emit()
        else:
            if self.is_left_click:
                self.left_clicked.emit()
            else:
                self.right_clicked.emit()

        self.is_double = False


class QLabelClickable(QLabel):
    # doubleClicked = pyqtSignal(object)
    clicked = pyqtSignal(object)

    def __init__(self, parent=None):
        super(QLabelClickable, self).__init__(parent)

    def mousePressEvent(self, event):
        self.ultimo = 'Click'

    def mouseReleaseEvent(self, event):
        if self.ultimo == 'Click':
            QTimer.singleShot(QApplication.instance().doubleClickInterval(),
                              self.performSingleClickAction)
        else:
            self.ultimo = 'doubleClick'
            self.clicked.emit(self.ultimo)

    def mouseDoubleClickEvent(self, event):
        self.ultimo = 'doubleClick'

    def performSingleClickAction(self):
        if self.ultimo == 'Click':
            self.clicked.emit(self.ultimo)
            self.update()


class QFrameClickable(QFrame):
    clicked = pyqtSignal(object)

    def __init__(self, parent=None):
        super(QFrameClickable, self).__init__(parent)

    # def mousePressEvent(self, QMouseEvent):
    #     if QMouseEvent.button() == Qt.LeftButton:
    #         self.clicked.emit()
    #     elif QMouseEvent.button() == Qt.RightButton:
    #         self.clicked.emit()
    #     self.update()

    def mousePressEvent(self, event):
        self.ultimo = 'Click'

    def mouseReleaseEvent(self, event):
        if self.ultimo == 'Click':
            QTimer.singleShot(QApplication.instance().doubleClickInterval(),
                              self.performSingleClickAction)
        else:
            self.ultimo = 'doubleClick'
            self.clicked.emit(self.ultimo)

    def mouseDoubleClickEvent(self, event):
        self.ultimo = 'doubleClick'

    def performSingleClickAction(self):
        if self.ultimo == 'Click':
            self.clicked.emit(self.ultimo)
            self.update()


class LineEditEntered(QLineEdit):
    sig_return_pressed = pyqtSignal()

    def __init__(self):
        QLineEdit.__init__(self)

    def keyPressEvent(self, event):
        super(LineEditEntered, self).keyPressEvent(event)

        if event.key() == Qt.Key_Return:
            self.sig_return_pressed.emit()
        elif event.key() == Qt.Key_Enter:
            self.sig_return_pressed.emit()



"""
@author: Jingyi GF{jingyi.g.fuglstad@gmail.com}
"""
import os
import sys
from sys import platform
from setuptools import setup, find_packages


is_problematic = False

if sys.version_info[:2] < (3, 8) or sys.version_info[:2] > (3, 11):
    is_problematic = True

if sys.version_info.minor == 8 and sys.version_info.micro < 10:
    is_problematic = True

if sys.version_info.minor == 9 and sys.version_info.micro == 0:
    is_problematic = True

if sys.version_info.minor == 10 and sys.version_info.micro >= 5:
    is_problematic = True

if is_problematic:
    raise RuntimeError("Python version >= 3.8.10 < 3.10.5 / 3.9.0 required.")

# Utility function to read the README file.
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


CLASSIFIERS = """
Development Status :: 3 - Alpha
Intended Audience :: Science/Research
Intended Audience :: Developers
Intended Audience :: Education
License :: OSI Approved :: MIT License
Programming Language :: Python :: 3
Programming Language :: Python :: 3.8
Programming Language :: Python :: 3.9
Programming Language :: Python :: 3.10
Programming Language :: Python :: 3 :: Only
Topic :: Software Development
Topic :: Scientific/Engineering
Typing :: Typed
Operating System :: Microsoft :: Windows
Operating System :: POSIX
Operating System :: Unix
Operating System :: MacOS
"""

REQUIRES = """
PyQt5 == 5.14.2; python_version == "3.8"
PyQt5 >= 5.15.1; python_version >= "3.9"
pyqtgraph >= 0.12.3
PyOpenGL >= 3.1.5
QtRangeSlider == 0.1.5
opencv-python >= 4.5.4.60
numba >= 0.54.1
numpy >= 1.20.3
scipy >= 1.7.3
requests >= 2.26.0
nibabel >= 3.2.1
pynrrd >= 0.4.3
tifffile >= 2021.11.2
aicspylibczi >= 3.0.3
pandas >= 1.3.5
natsort >= 8.0.2
imagecodecs >= 2022.2.22
h5py >= 3.7.0
tables >= 3.7.0
"""


# requires_py8 = """
# PyQt5 == 5.14.2
# pyqtgraph >= 0.12.3
# PyOpenGL >= 3.1.5
# QtRangeSlider == 0.1.5
# opencv-python >= 4.5.4.60
# numba >= 0.54.1
# numpy >= 1.20.3
# scipy >= 1.7.3
# requests >= 2.26.0
# nibabel >= 3.2.1
# pynrrd >= 0.4.3
# tifffile >= 2021.11.2
# aicspylibczi >= 3.0.3
# pandas >= 1.3.5
# natsort >= 8.0.2
# imagecodecs >= 2022.2.22
# h5py >= 3.7.0
# tables >= 3.7.0
# """

# if sys.version_info.minor == 8:
#     REQUIRES = requires_py8
# else:
#     REQUIRES = requires_py9

# if platform == "linux" or platform == "linux2":
#     # linux
# elif platform == "darwin":
#
# elif platform == "win32":



PACKAGE_DATA = """
main_window.ui
data/query.csv
data/atlas_labels.pkl
icons/*.svg
icons/*.png
icons/layers/*.svg
icons/layers/*.png
icons/sidebar/*.svg
icons/sidebar/*.png
icons/toolbar/*.svg
icons/toolbar/*.png
"""


setup(
    name="herbs",
    version="0.0.13",
    author="Jingyi GF",
    author_email="jingyi.g.fuglstad@gmail.com",
    description="A Python-based GUI for Histological E-data Registration in Brain Space",
    keywords="brain atlas, histological image registration, probe coordinates",
    url="https://github.com/JingyiGF/HERBS",
    packages=find_packages(),
    package_data={'': [_f for _f in PACKAGE_DATA.split('\n') if _f]},
    include_package_data=True,
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    project_urls={
        "Bug Tracker": "https://github.com/JingyiGF/HERBS/issues",
    },
    classifiers=[_f for _f in CLASSIFIERS.split('\n') if _f],
    python_requires=">=3.8",
    install_requires=[_f for _f in REQUIRES.split('\n') if _f]
)



import colorsys
import os
import sys

import PyQt5.QtGui
import cv2
import copy
import pyqtgraph.functions as fn
import numpy as np
import pyqtgraph as pg
from PyQt5.QtWidgets import *
from PyQt5.QtGui import *
from PyQt5.QtCore import *
from pyqtgraph.Qt import QtGui, QtCore
import scipy.ndimage as ndi

from .image_stacks import ImageStacks
from .image_curves import CurveWidget, ChannelSelector
from .uuuuuu import hsv2rgb, gamma_line, color_img, make_color_lut, get_corner_line_from_rect, \
    rotate, rotate_bound, get_tb_size


sidebar_button_style = '''
QPushButton{
    background: #656565;
    border-radius: 5px;
    color: white;
    border-style: outset;
    border-bottom: 1px solid rgb(30, 30, 30);
    height: 24px;
    margin: 0px;
}

QPushButton:hover{
    background-color: #323232;
    border: 1px solid #656565;
}

QPushButton:checked{
    background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
                                      stop: 0 #17bb00, stop: 1 #0e6900);
    border-bottom: 1px solid rgb(30, 30, 30);
    border-top: None;
    border-left: None;
    border-right: None;  
}
'''


class ImagePageController(QWidget):
    class SignalProxy(QObject):
        sigPageChanged = pyqtSignal(object)

    def __init__(self):
        self._sigprox = ImagePageController.SignalProxy()
        self.sig_page_changed = self._sigprox.sigPageChanged

        QWidget.__init__(self)

        # self.setStyleSheet(page_control_style)

        self.max_val = None

        self.page_slider = QSlider(Qt.Horizontal)
        self.page_slider.setMinimum(0)
        self.page_slider.valueChanged.connect(self.slider_value_changed)

        self.page_label = QLabel()
        self.page_label.setFixedSize(50, 20)

        self.page_left_btn = QPushButton()
        self.page_left_btn.setIcon(QtGui.QIcon("icons/backward.svg"))
        self.page_left_btn.setIconSize(QtCore.QSize(16, 16))
        self.page_left_btn.clicked.connect(self.left_btn_clicked)

        self.page_right_btn = QPushButton()
        self.page_right_btn.setIcon(QtGui.QIcon("icons/forward.svg"))
        self.page_right_btn.setIconSize(QtCore.QSize(16, 16))
        self.page_right_btn.clicked.connect(self.right_btn_clicked)

        page_ctrl_layout = QHBoxLayout()
        page_ctrl_layout.setSpacing(0)
        page_ctrl_layout.setContentsMargins(10, 5, 10, 5)
        page_ctrl_layout.addWidget(self.page_left_btn)
        page_ctrl_layout.addSpacing(10)
        page_ctrl_layout.addWidget(self.page_slider)
        page_ctrl_layout.addSpacing(5)
        page_ctrl_layout.addWidget(self.page_label)
        page_ctrl_layout.addWidget(self.page_right_btn)

        self.setLayout(page_ctrl_layout)

    def set_val(self, val):
        self.page_slider.setValue(val)

    def set_max(self, val):
        self.max_val = val
        self.page_slider.setMaximum(val)

    def slider_value_changed(self):
        val = self.page_slider.value()
        self.page_label.setText(str(val))
        self.sig_page_changed.emit(val)

    def left_btn_clicked(self):
        val = self.page_slider.value() - 1
        if val < 0:
            val = 0
        self.set_val(val)

    def right_btn_clicked(self):
        val = self.page_slider.value() + 1
        if val > self.max_val:
            val = self.max_val
        self.set_val(val)


class ImageView(QObject):
    """
    A collection of user interface elements bound together:
        for histological images usage.

    When set data into this object, the full image file is needed.
    * the maximum channel number is 4, matching ZenBlue.
    * scene_wrap, scale_wrap, img_stacks, curve_widget, chn_widget_wrap
    * if image file is rgb, chn_widget_wrap will not show
    * curve change for rgb ---> hsv on brightness channel
    * curve change for not rgb ----> on the channel which is visible


    """
    class SignalProxy(QObject):
        imageChanged = pyqtSignal()  # id

    def __init__(self):
        self._sigprox = ImageView.SignalProxy()
        self.sig_image_changed = self._sigprox.imageChanged

        QObject.__init__(self)

        # define the initials
        self.image_file = None
        self.processing_img = None
        self.current_img = None
        self.current_scale = None
        self.img_size = None
        self.tb_size = None
        self.color_lut_list = []
        self.original_lut_list = []
        self.scene_index = 0
        self.max_num_channels = 4
        self.channel_visible = [False, False, False, False]
        self.channel_color = [None, None, None, None]
        self.color_combo_index = [-1, -1, -1, -1]

        self.side_lines = None
        self.corner_points = None
        self.volume_img = None
        self.display_img_index = 0

        # scene control
        self.check_scenes = QPushButton('Load ALL Scenes')
        self.check_scenes.setStyleSheet(sidebar_button_style)
        self.check_scenes.setCheckable(True)
        self.scene_slider = QSlider(Qt.Horizontal)
        self.scene_slider.setMinimum(0)
        self.scene_slider.setValue(0)
        self.scene_slider.setTickPosition(QSlider.TicksBelow)
        self.scene_slider.setTickInterval(1)
        self.scene_slider.valueChanged.connect(self.scene_index_changed)
        self.scene_label = QLabel('0/ 0')
        self.scene_wrap = QFrame()
        # self.scene_wrap.setFixedWidth(290)
        scene_wrap_layout = QHBoxLayout(self.scene_wrap)
        # scene_wrap_layout.setContentsMargins(0, 0, 0, 0)
        # scene_wrap_layout.setSpacing(5)
        scene_wrap_layout.addWidget(QLabel('Scene: '))
        scene_wrap_layout.addWidget(self.scene_slider)
        scene_wrap_layout.addWidget(self.scene_label)
        # self.scene_wrap.setVisible(False)

        # scale control
        self.scale_slider = QSlider(Qt.Horizontal)
        self.scale_slider.setMinimum(1)
        self.scale_slider.setMaximum(100)
        self.scale_slider.setValue(10)
        self.scale_slider.setSingleStep(10)
        self.scale_slider.setTickInterval(10)
        self.scale_slider.valueChanged.connect(self.scale_value_changed)
        self.scale_slider.sliderReleased.connect(self.scale_slider_released)
        self.scale_label = QLabel('{}%'.format(10))
        self.scale_wrap = QFrame()
        scale_wrap_layout = QHBoxLayout(self.scale_wrap)
        scale_wrap_layout.addWidget(QLabel('Scale: '))
        scale_wrap_layout.addWidget(self.scale_slider)
        scale_wrap_layout.addWidget(self.scale_label)

        # image stacks
        self.img_stacks = ImageStacks()

        # scene control
        self.page_ctrl = ImagePageController()
        self.page_ctrl.setVisible(False)
        self.page_ctrl.sig_page_changed.connect(self.image_page_changed)

        # curve widget
        self.curve_widget = CurveWidget()
        self.curve_widget.setEnabled(False)
        self.curve_widget.setFixedHeight(280)
        self.curve_widget.sig_table_changed.connect(self.image_curve_changed)
        self.curve_widget.sig_reset.connect(self.image_curve_reset)
        self.curve_widget.sig_line_type_changed.connect(self.image_curve_type_changed)

        # channel buttons
        self.chn_widget_wrap = QFrame()
        self.chn_widget_wrap.setStyleSheet('QFrame{border: 1px solid #747a80; border-radius: 5px;}')
        chn_widget_layout = QHBoxLayout(self.chn_widget_wrap)
        self.chn_widget_list = []
        for i in range(self.max_num_channels):
            self.chn_widget_list.append(ChannelSelector())
            self.chn_widget_list[i].set_channel_index(i)
            self.chn_widget_list[i].sig_vis_channels.connect(self.set_channel_visible)
            self.chn_widget_list[i].sig_change_color.connect(self.channel_color_changed)
            chn_widget_layout.addWidget(self.chn_widget_list[i])
            self.chn_widget_list[i].setVisible(False)
        # self.chn_widget_wrap.setVisible(False)

        self.outer_frame = QFrame()
        outer_layout = QVBoxLayout(self.outer_frame)
        outer_layout.setContentsMargins(0, 0, 0, 0)
        outer_layout.setSpacing(0)
        outer_layout.addWidget(self.check_scenes)
        outer_layout.addWidget(self.scale_wrap)
        outer_layout.addWidget(self.scene_wrap)
        outer_layout.addWidget(self.curve_widget)
        outer_layout.addSpacing(10)
        outer_layout.addWidget(self.chn_widget_wrap)

    def set_data(self, image_file):
        if self.image_file is not None:
            self.page_ctrl.setVisible(False)
            self.clear_image_stacks()
            self.curve_widget.reset_pressed()
            self.curve_widget.setEnabled(False)
            for i in range(self.image_file.n_channels):
                self.chn_widget_list[i].delete_item()
            for i in range(self.max_num_channels):
                self.channel_color[i] = None
                self.channel_visible[i] = False
                self.img_stacks.image_list[i].clear()
                self.chn_widget_list[i].setVisible(False)
            self.image_file = None
            self.color_lut_list.clear()
            self.original_lut_list.clear()
            self.curve_widget.line_type_combo.setCurrentIndex(0)

        self.image_file = image_file

        if self.image_file.n_pages > 1:
            self.page_ctrl.set_max(self.image_file.n_pages - 1)
            self.page_ctrl.setVisible(True)
            self.volume_img = copy.deepcopy(self.image_file.data['scene 0'])
            init_page_number = int(0.5 * self.image_file.n_pages)
            self.page_ctrl.set_val(init_page_number)
            temp_data = self.volume_img[init_page_number, :, :].copy()
            temp_data = [temp_data]
            self.current_img = np.dstack(temp_data)
        else:
            self.volume_img = None
            self.current_img = copy.deepcopy(self.image_file.data['scene 0'])

        self.current_scale = self.image_file.scale['scene 0']

        if self.image_file.n_scenes != 1:
            self.scene_slider.setMaximum(self.image_file.n_scenes - 1)
            self.scene_label.setText('1/{}'.format(self.image_file.n_scenes))
            self.scene_wrap.setVisible(True)
        else:
            self.scene_wrap.setVisible(False)

        if not self.image_file.is_czi:
            self.scale_wrap.setVisible(False)
        else:
            self.scale_wrap.setVisible(True)

        # set data to curves
        self.set_curve_widgets()  # get reasonable table
        self.set_channel_widgets()  # get initial color_lut_list for each layer

        # set data to image stacks
        self.set_data_to_img_stacks()
        self.get_corner_and_lines()

    def set_channel_widgets(self):
        # set color and names to channels
        self.chn_widget_wrap.setVisible(True)
        for i in range(self.image_file.n_channels):
            # get color look-up-table
            self.channel_color[i] = copy.deepcopy(self.image_file.hsv_colors[i])
            da_hsv_color = self.channel_color[i]
            da_lut = make_color_lut(da_hsv_color, self.image_file.level + 1)
            self.original_lut_list.append(da_lut)
            self.color_lut_list.append(da_lut)
            self.chn_widget_list[i].setVisible(True)
            self.chn_widget_list[i].vis_btn.setText(self.image_file.channel_name[i])
            self.chn_widget_list[i].add_item(self.channel_color[i])
            self.channel_visible[i] = True
            self.color_combo_index[i] = self.chn_widget_list[i].color_combo.currentIndex()

    def set_curve_widgets(self):
        self.curve_widget.setEnabled(True)
        img_layers = self.current_img.copy()
        self.curve_widget.set_data(img_layers, self.image_file.rgb_colors, self.image_file.level)

    def channel_color_changed(self, col, ind):
        da_lut = make_color_lut(col, self.curve_widget.gray_max + 1)
        self.original_lut_list[ind] = da_lut.copy()
        self.color_lut_list[ind] = da_lut[self.curve_widget.table_output[ind], :]
        self.img_stacks.image_list[ind].setLevels(levels=(0, self.image_file.level))
        self.img_stacks.image_list[ind].setLookupTable(self.color_lut_list[ind])
        self.channel_color[ind] = col
        self.curve_widget.curve_plot.change_hist_color(col, ind)
        self.color_combo_index[ind] = self.chn_widget_list[ind].color_combo.currentIndex()

    def set_channel_visible(self, vis, ind):
        if np.sum(self.channel_visible) == 1:
            vis_layer_ind = np.where(np.ravel(self.channel_visible))[0][0]
            if vis_layer_ind == ind:
                return
        self.channel_visible[ind] = vis
        self.img_stacks.image_list[ind].setVisible(vis)
        self.curve_widget.set_channel_enable(ind, vis)

    def channel_selector_reset(self):
        for i in range(self.image_file.n_channels):
            self.chn_widget_list[i].vis_btn.setChecked(False)
            self.chn_widget_list[i].vis = True
            self.chn_widget_list[i].set_checked(False)

    def set_data_and_size(self, img_data):
        self.img_stacks.set_data(img_data)
        self.img_size = img_data.shape[:2]
        rect = (0, 0, self.img_size[1], self.img_size[0])
        self.corner_points, self.side_lines = get_corner_line_from_rect(rect)
        self.img_stacks.image_dict['tri_pnts'].set_range(self.img_size[1], self.img_size[0])
        self.tb_size = get_tb_size(self.img_size)

    def set_data_to_img_stacks(self):
        self.set_data_and_size(self.current_img)
        self.img_stacks.set_lut(self.color_lut_list, self.curve_widget.gray_max)

    def get_corner_and_lines(self):
        rect = (0, 0, self.img_size[1], self.img_size[0])
        self.corner_points, self.side_lines = get_corner_line_from_rect(rect)
        self.sig_image_changed.emit()

    def image_page_changed(self, page_number):
        self.display_img_index = page_number
        temp_data = self.volume_img[page_number, :, :].copy()
        temp_data = [temp_data]
        self.current_img = np.dstack(temp_data)
        self.img_stacks.set_data(self.current_img)

    def scene_index_changed(self):
        if self.image_file is None:
            return
        scene_index = self.scene_slider.value()
        self.scene_label.setText('{}/{}'.format(scene_index + 1, self.image_file.n_scenes))
        with pg.BusyCursor():
            self.clear_image_stacks()
            if 'scene {}'.format(scene_index) not in list(self.image_file.data.keys()):
                scale = self.scale_slider.value()
                scale_val = scale * 0.01
                self.image_file.read_data(scale_val, scene_index=scene_index)
            self.current_img = copy.deepcopy(self.image_file.data['scene {}'.format(scene_index)])
            self.current_scale = self.image_file.scale['scene {}'.format(scene_index)]

            if self.current_scale != self.scale_slider.value() * 0.01:
                self.scale_slider.setValue(self.current_scale * 100)

            self.img_size = self.current_img.shape[:2]
            for i in range(self.image_file.n_channels):
                self.img_stacks.image_list[i].clear()
            self.img_stacks.image_dict['tri_pnts'].set_range(self.img_size[1], self.img_size[0])

            self.channel_selector_reset()
            self.curve_widget.curve_plot.reset_plot()

            self.img_stacks.set_data(self.current_img)
            self.img_stacks.set_lut(self.original_lut_list, self.image_file.level)
            self.get_corner_and_lines()
            self.set_curve_widgets()
            # reset gamma and channels and black/white slider

    def scale_value_changed(self):
        scale = self.scale_slider.value()
        self.scale_label.setText('{}%'.format(scale))

    def scale_slider_released(self):
        if self.image_file is None:
            return
        if self.image_file.is_czi:
            self.clear_image_stacks()
            scale = self.scale_slider.value()
            scale_val = scale * 0.01
            scene_index = self.scene_slider.value()
            with pg.BusyCursor():
                self.image_file.read_data(scale_val, scene_index)
                self.current_img = copy.deepcopy(self.image_file.data['scene %d' % scene_index])
                self.current_scale = self.image_file.scale['scene {}'.format(scene_index)]

                self.img_size = self.current_img.shape[:2]
                for i in range(self.image_file.n_channels):
                    self.img_stacks.image_list[i].clear()
                self.img_stacks.image_dict['tri_pnts'].set_range(self.img_size[1], self.img_size[0])
                self.set_data_to_img_stacks()
                self.get_corner_and_lines()

    def image_curve_changed(self, ev):
        if self.current_img is None:
            return
        table = ev[0]
        ind = ev[1]
        lut_in = self.original_lut_list[ind].copy()
        self.color_lut_list[ind] = lut_in[table, :]
        self.img_stacks.image_list[ind].setLookupTable(self.color_lut_list[ind])
        # table = table.astype(int)
        # if self.image_file.pixel_type == 'rgb24':
        #     da_img = cv2.LUT(self.current_img.astype('uint8'), table)
        #     self.img_stacks.set_data(da_img)
        # else:
        #     for i in range(self.image_file.n_channels):
        #         if self.channel_visible[i]:
        #             lut_in = self.color_lut_list[i].copy()
        #             self.color_lut_list[i] = lut_in[table, :]
        #             self.img_stacks.image_list[i].setLookupTable(self.color_lut_list[i])

    def image_curve_type_changed(self):
        """
        signal function after change line type
        set to the original 2 points linear type
        """
        if self.image_file is None:
            return
        for i in range(self.image_file.n_channels):
            self.img_stacks.image_list[i].setLookupTable(self.original_lut_list[i])

    def image_curve_reset(self):
        """
        signal function after press button 'Reset'
        """
        for i in range(self.image_file.n_channels):
            self.channel_color[i] = self.image_file.hsv_colors[i]
        for i in range(self.image_file.n_channels):
            self.chn_widget_list[i].color_combo.setCurrentIndex(
                len(self.chn_widget_list[i].color_combo.hsv_color_list) - 1)
            self.curve_widget.curve_plot.change_hist_color(self.channel_color[i], i)
            self.img_stacks.image_list[i].setLookupTable(self.original_lut_list[i])
        self.color_lut_list = self.original_lut_list.copy()

    def get_img_ctrl_data(self):
        lut_points_data = []
        for i in range(self.current_img.shape[2]):
            lut_points_data.append(self.curve_widget.curve_plot.lut_points[i].data['pos'].copy())

        data = {'current_img': self.current_img,
                'processing_img': self.processing_img,
                'current_scene': self.scene_slider.value(),
                'current_scale': self.scale_slider.value(),
                'channel_color': self.channel_color,
                'color_combo_index': self.color_combo_index,
                'original_lut_list': self.original_lut_list,
                'color_lut_list': self.color_lut_list,
                'gamma_vals': self.curve_widget.gamma,
                'table_output': self.curve_widget.table_output,
                'original_table': self.curve_widget.original_table,
                'line_type': self.curve_widget.line_type,
                'hist_data': self.curve_widget.curve_plot.hist_data,
                'lut_points_data': lut_points_data}
        return data

    def load_img_ctrl_data(self, img_ctrl_data):
        self.current_img = img_ctrl_data['current_img']
        self.current_scale = img_ctrl_data['current_scale']
        self.channel_color = img_ctrl_data['channel_color']
        self.color_combo_index = img_ctrl_data['color_combo_index']
        self.processing_img = img_ctrl_data['processing_img']
        # set data
        # self.scene_wrap.setVisible(False)

        self.scene_slider.blockSignals(True)
        self.scene_slider.setValue(img_ctrl_data['current_scene'])
        self.scene_label.setText('{}/{}'.format(img_ctrl_data['current_scene'] + 1, self.image_file.n_scenes))
        self.scene_slider.blockSignals(False)

        self.curve_widget.gamma = img_ctrl_data['gamma_vals']
        self.curve_widget.table_output = img_ctrl_data['table_output']
        self.curve_widget.original_table = img_ctrl_data['original_table']
        self.curve_widget.line_type = img_ctrl_data['line_type']

        self.curve_widget.line_type_combo.blockSignals(True)
        self.curve_widget.line_type_combo.setCurrentText(self.curve_widget.line_type)
        self.curve_widget.line_type_combo.blockSignals(False)

        self.curve_widget.curve_plot.line_type = self.curve_widget.line_type
        self.curve_widget.curve_plot.set_data(img_ctrl_data['hist_data'], self.image_file.rgb_colors,
                                              self.image_file.level)

        for i in range(self.image_file.n_channels):
            self.curve_widget.curve_plot.change_hist_color(self.channel_color[i], i)
            self.curve_widget.curve_plot.set_lut_line(self.curve_widget.table_output[i], i)
            self.curve_widget.curve_plot.set_lut_points(img_ctrl_data['lut_points_data'][i], i)

        for i in range(self.image_file.n_channels):
            self.curve_widget.curve_plot.set_enable(i, True)
        self.curve_widget.set_enable_induced_slider()

        self.original_lut_list = img_ctrl_data['original_lut_list']
        self.color_lut_list = img_ctrl_data['color_lut_list']

        for i in range(self.image_file.n_channels):
            self.chn_widget_list[i].blockSignals(True)
            self.chn_widget_list[i].setVisible(True)
            self.chn_widget_list[i].vis_btn.setChecked(False)
            self.chn_widget_list[i].set_checked(False)
            self.chn_widget_list[i].vis_btn.setText(self.image_file.channel_name[i])
            self.chn_widget_list[i].color_combo.blockSignals(True)
            self.chn_widget_list[i].add_item(self.image_file.hsv_colors[i])
            self.chn_widget_list[i].color_combo.setCurrentIndex(self.color_combo_index[i])
            self.chn_widget_list[i].color_combo.blockSignals(False)
            self.channel_visible[i] = True
            self.chn_widget_list[i].blockSignals(False)

        self.original_lut_list = img_ctrl_data['original_lut_list']
        self.color_lut_list = img_ctrl_data['color_lut_list']

        # self.set_data_and_size(self.processing_img)

    #
    def check_img_process_layer_data(self, layer_dict):
        valid_keys = list(layer_dict.keys())
        check_res = True
        if not np.all(['data', 'level', 'size'] in valid_keys):
            check_res = False
        if not np.all(layer_dict['size'] == self.image_view.current_img.shape[:2]):
            check_res = False
        if layer_dict['level'] != self.image_file.level:
            check_res = False
        return check_res

    def has_loaded_layer_the_same_size(self, layer_dict):
        check_res = True
        if not np.all(layer_dict['data'].shape[:2] == self.image_view.current_img.shape[:2]):
            check_res = False
        return check_res

    def save_img_process_data(self):
        data = {'data': self.processing_img,
                'level': self.image_file.level,
                'size': self.img_size}
        return data

    # clear image stacks
    def clear_image_stacks(self):
        valid_keys = ['img-overlay', 'overlay_contour', 'img-mask', 'circle_follow', 'lasso_path', 'img-virus',
                      'img-contour', 'img-probe', 'img-cells', 'img-blob', 'img-drawing']

        for da_key in valid_keys:
            self.img_stacks.image_dict[da_key].clear()

    # image rotation
    def image_vertical_flip(self):
        if self.image_file is None:
            return
        # if self.image_file.is_rgb:
        #     self.current_img = cv2.flip(self.current_img, 0)
        # else:
        for i in range(self.image_file.n_channels):
            self.current_img[:, :, i] = cv2.flip(self.current_img[:, :, i], 0)
        self.clear_image_stacks()
        self.img_size = self.current_img.shape[:2]
        self.img_stacks.image_dict['tri_pnts'].set_range(self.img_size[1], self.img_size[0])
        self.img_stacks.set_data(self.current_img)

    def image_horizon_flip(self):
        if self.image_file is None:
            return
        # if self.image_file.is_rgb:
        #     self.current_img = cv2.flip(self.current_img, 1)
        # else:
        for i in range(self.image_file.n_channels):
            self.current_img[:, :, i] = cv2.flip(self.current_img[:, :, i], 1)
        self.clear_image_stacks()
        self.img_stacks.set_data(self.current_img)
        self.img_size = self.current_img.shape[:2]
        self.img_stacks.image_dict['tri_pnts'].set_range(self.img_size[1], self.img_size[0])

    def image_90_rotate(self):
        if self.image_file is None:
            return
        temp = []
        for i in range(self.image_file.n_channels):
            self.img_stacks.image_list[i].clear()
            temp.append(cv2.rotate(self.current_img[:, :, i], cv2.ROTATE_90_CLOCKWISE))
        self.current_img = np.dstack(temp)
        self.clear_image_stacks()
        self.img_stacks.set_data(self.current_img)
        self.img_size = self.current_img.shape[:2]
        self.img_stacks.image_dict['tri_pnts'].set_range(self.img_size[1], self.img_size[0])
        self.get_corner_and_lines()
        # self.image_file.data['scene %d' % scene_index] = self.current_img.copy()
        self.tb_size = np.flip(self.tb_size, 0)

    def image_180_rotate(self):
        if self.image_file is None:
            return
        # if self.image_file.is_rgb:
        #     self.current_img = cv2.rotate(self.current_img, cv2.ROTATE_180)
        # else:
        temp = []
        for i in range(self.image_file.n_channels):
            self.img_stacks.image_list[i].clear()
            temp.append(cv2.rotate(self.current_img[:, :, i], cv2.ROTATE_180))
        self.current_img = np.dstack(temp)
        self.clear_image_stacks()
        self.img_stacks.set_data(self.current_img)
        self.img_size = self.current_img.shape[:2]
        self.img_stacks.image_dict['tri_pnts'].set_range(self.img_size[1], self.img_size[0])

    def image_90_counter_rotate(self):
        if self.image_file is None:
            return
        # if self.image_file.is_rgb:
        #     self.current_img = cv2.rotate(self.current_img, cv2.ROTATE_90_COUNTERCLOCKWISE)
        # else:
        temp = []
        for i in range(self.image_file.n_channels):
            self.img_stacks.image_list[i].clear()
            temp.append(cv2.rotate(self.current_img[:, :, i], cv2.ROTATE_90_COUNTERCLOCKWISE))
        self.current_img = np.dstack(temp)
        self.clear_image_stacks()
        self.img_stacks.set_data(self.current_img)
        self.img_size = self.current_img.shape[:2]
        self.img_stacks.image_dict['tri_pnts'].set_range(self.img_size[1], self.img_size[0])
        self.get_corner_and_lines()
        self.tb_size = np.flip(self.tb_size, 0)

    def image_1_rotate(self, rotate_direction):
        if self.image_file is None:
            return
        if rotate_direction == 'clockwise':
            rotation_angle = - 1
        else:
            rotation_angle = 1
        temp = []
        for i in range(self.image_file.n_channels):
            temp.append(rotate(self.current_img[:, :, i], rotation_angle))
        self.current_img = np.dstack(temp)
        self.clear_image_stacks()
        self.img_stacks.set_data(self.current_img)
        self.img_size = self.current_img.shape[:2]
        self.img_stacks.image_dict['tri_pnts'].set_range(self.img_size[1], self.img_size[0])
        self.get_corner_and_lines()

    # def clear_curve_widget(self):




